

from scipy.special import erf
import numpy as np
import scipy.integrate


def gauss2d(p, x, y, u, x0=0.0, y0=0.0):
    """\
    Computes Gaussian plume (units: kg/m2).

    p: state vector (Q, K, BG)
    - Q:  emissions in kg/s
    - K:  eddy diffusion coefficient in m2/s
    - BG: background field in kg/m2

    x: distance from origin in m
    y: distance from center line in m
    u: wind speed in m/s

    x0: x-coordinate of origin
    y0: y-coordinate of origin

    TODO:
    - decay time

    """
    # state vector
    Q = float(p[0])    # emissions (kg/s)
    K = float(p[1])    # eddy diffusion coefficient (m2 / s)

    if len(p) == 3:
        BG = float(p[2])   # background field (kg/m2)
    else:
        BG = 0.0

    # dispersion along the plume
    sigma = np.sqrt(2.0 * K * (x-x0) / u)

    # compute Gaussian plume
    A = Q / (np.sqrt(2.0 * np.pi) * sigma * u) * np.exp(-0.5 * (y-y0)**2 / sigma**2)

    plume = np.full(x.shape, BG)
    plume[x > 0.0] += A[x > 0.0]

    return plume



def gauss2d_integral(x, lower, upper):
    """
    Integral form of gauss function over y.
    """
    if x <= 0:
        raise ValueError

    x0 = 0.0
    K = 3000.0

    Q = 600.0
    u = 10.0

    sigma = np.sqrt(2.0 * K * (x-x0) / u)

    xu = np.polyval(upper, x) / np.sqrt(2) / sigma
    xl = np.polyval(lower, x) / np.sqrt(2) / sigma

    return Q / u * erf(xu) - erf(xl)




def compute_line(x,y, k,l):
    """
    Compute slope and intercept between x[k] and x[l].
    """
    slope = (y[l] - y[k]) / (x[l] - x[k])
    intercept = y[k] - slope * x[k]
    return slope, intercept


def iter_regions(x,y):
    """
    Compute lower/upper line (+1/-1).

    Returns

    xt[0], xt[1], m, b
    xt[1], xt[2], m, b
    ...
    """
    start = np.argmin(x)
    xt = np.sort(x)

    # upper line (anticlockwise: -1)
    k_up = start
    l_up = (start - 1) % x.size

    # slope and intercept
    c_up = compute_line(x, y, k_up, l_up)

    # lower line (clockwise: +1)
    k_low = start
    l_low = (start + 1) % x.size

    # slope and intercept
    c_low = compute_line(x, y, k_low, l_low)

    for i in range(xt.size-1):

        # update upper line
        if xt[i] > x[l_up]:
            k_up = (k_up - 1) % x.size
            l_up = (k_up - 1) % x.size
            c_up = compute_line(x, y, k_up, l_up)

        # update lower line
        if xt[i] > x[l_low]:
            k_low = (k_low + 1) % x.size
            l_low = (k_low + 1) % x.size
            c_low = compute_line(x, y, k_low, l_low)

        yield xt[i], xt[i+1], c_low, c_up


def compute_mean(x,y):
    """
    Compute mean of Gaussian over convex polygon given by xc and yc.
    """
    total = 0.0

    # remove negative values (integral is zero)
    if np.all(x <= 0.0):
        return total

    if np.any(x <= 0.0):
        raise NotImplementedError('Some x-values are negative!')

    # sort left to right
    n = x.size

    sort = np.argsort(x)
    xt = x[sort]

    start = np.argmin(x)
    stop = np.argmax(x)

    for a, b, c_low, c_up in iter_regions(x,y):
        total += scipy.integrate.quad(
            gauss2d_integral, a, b, args=(c_low,c_up)[0]
        )

    return total / 2e3**2 # TODO: use area


def test(data):
    x = np.array(data.x)
    y = np.array(data.y)
    c = gauss2d([600.0, 3000.0], x, y, 10.0)

    # TODO: test and speed
    xc = np.array(data.xc)
    yc = np.array(data.yc)

    n,m,_ = data.xc.shape
    cm = np.full((n,m), np.nan)

    for i in range(n):
        for j in range(m):
            if data.plume_area[i,j]:
                cm[i,j] = compute_mean(xc[i,j], yc[i,j])


    return c, cm



