

"""
Module for estimating CO2/NO2 background from pixels surrounding a detected
plume.
"""

import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndi
import xarray as xr
import pykrige.kriging_tools as kt
from pykrige.ok import OrdinaryKriging
from pykrige.uk import UniversalKriging

import amrs
import emiss
import smartcarb.map as map


def create_gaussian_kernel(sigma):
    """
    Creat a Gaussian kernel with standard deviation `sigma`. The size of the
    kernel is at least 11x11 but at least 5*sigma.
    """

    # size should be odd and at least 11 pixels
    size = max(11, int(5 * sigma))
    if size%2 == 0:
        size += 1

    kernel = np.zeros((size,size))
    kernel[size//2,size//2] = 1.0
    kernel = ndi.gaussian_filter(kernel, sigma=sigma)

    return kernel



def create_input(data, tracer):
    """
    Create satellite data with gaps for elevated values and clouds.
    """
    if isinstance(tracer,str):
        values = data[tracer]
    c = np.array(values)

    area = np.asarray(data.plume_area)

    # only use pixels that are in plume area around plume without enhanced
    # tracer values
    valids = np.array(area & ~data.detected_plume & ~data.is_hit)
    valids[~np.isfinite(c)] = False

    c[~valids] = np.nan

    x = data.x.values[valids]
    y = data.y.values[valids]
    c = c[valids]

    return x, y, c


def estimate_by_kriging(data, tracer, model='OK', variogram_model='linear'):
    """
    Estimate background field in plume area using Kriging.

    linear, power, gaussian, spherical, exponential, hole-effect
    """

    x, y, c = create_input(data, tracer)

    area = data.plume_area
    px = data.x.values[area]
    py = data.y.values[area]

    est = np.full(data[tracer].shape, np.nan)
    var = np.full(data[tracer].shape, np.nan)

    # kriging
    if model == 'OK':
        model = OrdinaryKriging(x,y,c, variogram_model=variogram_model, verbose=False,
                                enable_plotting=False)
    else:
        model = UniversalKriging(x,y,c, variogram_model=variogram_model, verbose=False,
                                 enable_plotting=False)

    est[area], var[area] = model.execute('points', px, py)


    return est, var, model



def compare_with_true(bg_est, data, tracer):
    area = np.asarray(data.plume_area)
    bg_true = np.array(data['%s_noisefree' % tracer] -
                       data['%s_true_signal' % tracer])
    bg_true[~area] = np.nan

    # mean bias and standard deviation
    true_plume = data['detected_plume'].values & data['%s_isfinite' %
                                                      tracer]

    MB = np.nanmean((bg_est - bg_true)[true_plume])
    SD = np.nanstd((bg_est - bg_true)[true_plume])

    return MB, SD




def estimate(data, tracer, sigma=10.0, show=False, vmin=None, vmax=None):
    """
    Estimate (smooth) CO2 background using normalized convolution.
    """

    if isinstance(tracer,str):
        values = data[tracer]
    c = np.array(values)

    area = np.asarray(data.plume_area)

    # only use pixels that are in plume area around plume without enhanced
    # tracer values
    valids = np.array(area & ~data.detected_plume & ~data.is_hit)
    valids[~np.isfinite(c)] = False

    kernel = create_gaussian_kernel(sigma)

    c[~valids] = np.nan
    bg_est = amrs.misc.normalized_convolution(c, kernel, mask=~valids)
    bg_est[~area] = np.nan

    # true background and clouds
    if '%s_true_plume' % tracer in data:
        bg_true = np.array(data['%s_noisefree' % tracer] -
                           data['%s_true_signal' % tracer])
        bg_true[~area] = np.nan

        # mean bias and standard deviation
        true_plume = data['detected_plume'].values & data['%s_isfinite' %
                                                          tracer]

        MB = np.nanmean((bg_est - bg_true)[true_plume])
        SD = np.nanstd((bg_est - bg_true)[true_plume])
    else:
        bg_true = np.full(data[tracer].shape, np.nan)
        MB = np.nan
        SD = np.nan

    clouds = np.ma.array(data.clouds > 0.01, mask=~area)


    bg_est = xr.DataArray(bg_est, name='%s_estimated_background' % tracer,
                          dims=values.dims)
    bg_est.attrs['long name'] = 'estimated %s background' % tracer
    bg_est.attrs['method'] = 'normalized convolution (sigma = %.1f px)' % sigma
    bg_est.attrs['bias'] = MB
    bg_est.attrs['scatter'] = SD

    bg_true = xr.DataArray(bg_true, name='%s_true_background' % tracer,
                           dims=values.dims)
    bg_true.attrs['long name'] = 'true %s background' % tracer

    data['%s_estimated_background' % tracer] = bg_est
    data['%s_true_background' % tracer] = bg_true

    if show:
        if vmin is None:
            vmin = np.floor(np.nanmean(bg_est))

        if vmax is None:
            vmax = np.ceil(np.mean(bg_est))

        # estimated background
        if tracer == 'CO2':
            label = 'XCO$_2$ background (ppm)'
        else:
            label = 'NO$_2$ background (10^15 molec. cm-2)'
        fig1 = map.make_level2_map(data.lonc, data.latc, bg_est,
                                   vmin=vmin, vmax=vmax, do_zoom=True,
                                   clct=clouds, domain='smartcarb',
                                   label=label, fig_size=5, dright=0.01)

        fig1.canvas.set_window_title('%s estimated by normalized '\
                                     'convolution' % tracer)


        if '%s_true_plume' % tracer in data:
            fig2 = map.make_level2_map(data.lonc, data.latc, bg_true,
                                       label=label, fig_size=5, dright=0.01,
                                       clct=clouds, domain='smartcarb',
                                       do_zoom=True, vmin=vmin, vmax=vmax)

            fig2.canvas.set_window_title('True %s background' % tracer)


            delta = bg_est - bg_true
            delta.values[clouds] = np.nan
            fig3 = map.make_level2_map(data.lonc, data.latc, delta,
                                       label=label, fig_size=5, dright=0.01,
                                       domain='smartcarb',
                                       do_zoom=True, vmin=-1, vmax=+1,
                                       cmap='seismic')

            fig3.canvas.set_window_title('True %s background' % tracer)


        else:
            fig2 = None
            fig3 = None

        for fig in [fig1, fig2, fig3]:
            if fig is None:
                continue

            lines = map.add_scatter(fig.axes[0], data.lon, data.lat, ms=0.7,
                            slice=data.detected_plume, label='__none__')
            map.add_city_labels(fig.axes[0])

            if '%s_true_plume' in data:
                lines += amrs.vis.draw_contours(fig.axes[0], data.lon, data.lat,
                                                data['%s_true_plume' % tracer],
                                                label='%s plume' % tracer)

            # use figure instead of geoaxes because freeze with pcolormesh
            fig.legend(lines, [line.get_label() for line in lines],
                       loc=(0.025, 0.89), numpoints=1, markerscale=1.5,
                       fontsize='small')


    return data




def estimate_by_sigmas(filename):
    """
    """
    print(filename)
    sigmas = np.logspace(-1, 1, 10)

    data, curve = emiss.gauss.load_prepared(filename, co2_ssp='low',
                                            no2_ssp='high')
    bias = []
    scatter = []

    for sigma in sigmas:
        data = emiss.background.estimate(data, 'CO2', sigma=sigma, show=False)
        bias.append(data.CO2_estimated_background.bias)
        scatter.append(data.CO2_estimated_background.scatter)

    bias = np.array(bias)
    scatter = np.array(scatter)

    return sigmas, bias, scatter



def test_kriging(filenames):

    bias = []
    scatter = []

    for filename in filenames:
        print(filename)
        data, curve = emiss.gauss.load_prepared(filename, co2_ssp='low',
                                                no2_ssp='high')

        bg_est, bg_var, model = estimate_by_kriging(data, 'CO2', model='OK',
                                                    variogram_model='linear')

        mb, sd = compare_with_true(bg_est, data, 'CO2')

        bias.append(mb)
        scatter.append(sd)

    return np.array(bias), np.array(scatter)













