
"""
Compute CO2/NO2 mass along line or within area.
"""

from scipy.constants import N_A

from shapely.geometry import LineString, Polygon

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize

import doas
import amrs.vis

import smartcarb.map

from . import misc
from . import functions


def compute_mass(data, curve, field):
    """
    """
    distances = np.arange(0, np.nanmax(data.xp), 2e3)

    mass, mass_std = integrate_fields(data, field, curve, distances,
                                      tmax=100e3)

    valids = (distances > dmin) & np.isfinite(fluxes) & np.isfinite(fluxes_std)
    valids &= (distances <= dmax)

    mask = ~valids

    return np.ma.array(mass, mask=mask), np.ma.array(mass_std, mask=mask)



def integrate_fields(data, tracer, curve, distances, tmax):

    mass = []
    mass_std = []

    for d in distances:
        f,u = integrate_field(data, tracer, curve, d, tmax=tmax, show=False)
        mass.append(f)
        mass_std.append(u)

    return np.array(fluxes), np.array(fluxes_std)



def integrate_along_line(data, tracer, distances, curve=None, show=False,
                         method='sub-areas'):
    """
    distance from source origin
    """
    masses_true = []
    masses = []
    masses_std = []

    wind_eff = []
    wind_true = []

    for i in range(distances.size-1):

        results = extract_line_density(data, tracer, distances[i],
                                       distances[i+1],
                                       ya=None, yb=None, dy=5e3,
                                       method='both', show=False)

        masses.append(results[method])
        masses_std.append(results['%s_std' % method])

        masses_true.append(results['density_true'])

        wind_eff.append(results['wind'])
        wind_true.append(results['wind_true'])

    wind_true = np.array(wind_true)
    wind_eff = np.array(wind_eff)

    return np.array(masses), np.array(masses_std), np.array(masses_true), \
           wind_eff, wind_true




def add_area(ax, data, curve, distances, delta=10e3, color='yellow',
             add_label=True):

    # across-plume distance
    y = get_plume_width(data)

    # compute arc length for parameters: arc(t)
    t = np.linspace(0.0, curve.t_o + distances.max() + delta, 5001)
    arc = misc.compute_arc_length(curve, curve.t_o, t)

    for i in range(distances.size):

        # interpolate distance from arc(t)
        xmin, xmax = np.interp(
            [distances[i] - delta / 2, distances[i] + delta / 2],
            arc, t
        )

        # compute normals
        (x0, x1), (y0, y1) = curve.compute_normal(xmin, t=y)
        (x3, x2), (y3, y2) = curve.compute_normal(xmax, t=y)

        # plot box
        label = 'control polygons 'if add_label and (i == 0) else None
        xx,yy = np.array([x0, x1, x2, x3, x0]), np.array([y0, y1, y2, y3, y0])
        smartcarb.map.add_curve(ax, xx, yy, origin='Berlin', ls='-',
                                color=color, label=label)



def sort_and_remove_nans(y, c, c_std):
    """
    Sort for y and remove nans.
    """
    sort = np.argsort(y)
    y = y[sort]
    c = c[sort]
    c_std = c_std[sort]

    return y, c, c_std



def fit_two_gaussian(y1, c1, c_std1, y2, c2, c_std2, s=None):

    def co2_func(x, *a):
        """
        """
        return a[0] + a[1] * x + a[2] * np.exp(-4.0 * np.log(2) * (x - a[3])**2 /
                                               a[4]**2 )

    def no2_func(x, *a):
        return a[5] + a[6] * x + a[7] * np.exp(-4.0 * np.log(2) * (x - a[8])**2 /
                                               a[4]**2 )

    # sort and remove nans
    y1, c1, c_std1 = sort_and_remove_nans(y1, c1, c_std1)
    y2, c2, c_std2 = sort_and_remove_nans(y2, c2, c_std2)

    # concat
    y = np.concatenate([y1, y2])
    c = np.concatenate([c1, c2])


    def cost_function(x, *a):
        return np.concatenate([co2_func(y1, *a), no2_func(y2, *a)])

    p0 = np.array([np.mean(c1), 0.0, np.mean(c1), 0.0, 10e3,
                   np.mean(c2), 0.0, np.mean(c2), 0.0,       ])


    p, cov_p = scipy.optimize.curve_fit(cost_function, y, c, p0)

    co2_integral = 0.5 * np.sqrt( np.pi / np.log(2) ) * p[2] * p[4]
    no2_integral = 0.5 * np.sqrt( np.pi / np.log(2) ) * p[7] * p[4]

    return co2_integral, no2_integral, co2_func(s, *p), no2_func(s, *p)



def fit_function(y, c, c_std, s=None, function='gauss'):


    # sort
    sort = np.argsort(y)
    y = y[sort]
    c = c[sort]
    c_std = c_std[sort]

    # remove nans
    valids = np.isfinite(c)
    y = y[valids]
    c = c[valids]
    c_std = c_std[valids]

    if function in ['gaussian', 'gauss']:
        curve_function = functions.gauss
        p0 = np.array([
            100.0,  # integral
            10e3,   # sigma
            0.0     # shift
        ])

    elif function == 'erf':
        curve_function = functions.error_function
        p0 = np.array([
            100.0, # integral
            1e3,   # sigma
            20e3,  # box width
            0.0    # shift
        ])
    else:
        raise ValueError

    if np.all(np.isnan(c_std)):
        sigma = None
    else:
        sigma = c_std

    if y.size < p0.size:
        return np.nan, np.nan, np.full(s.shape, np.nan)

    try:
        p, cov_p = scipy.optimize.curve_fit(curve_function, y, c, p0, sigma=sigma,
                                            absolute_sigma=sigma is not None)
    except RuntimeError as e:
        return np.nan, np.nan, np.full(s.shape, np.nan)

    if function in ['gaussian', 'gauss',  'erf']:
        integral = p[0]
        integral_std = np.sqrt(cov_p[0,0])
    else:
        raise ValueError

    return integral, integral_std, curve_function(s, *p)



def extract_pixels(data, tracer, xa, xb, ya, yb, dy=None, get_wind=False):
    """
    Extract pixel coords and values for tracer in area.
    """

    # only use pixels that are valid observations
    xp = data['xp'].values
    yp = data['yp'].values

    area = (xa <= xp) & (xp < xb) & (ya <= yp) & (yp <= yb)

    isfinite = data['%s_isfinite' % tracer[:3]].values[area]

    # pixel in area
    xx = data.xp.values[area]
    yy = data.yp.values[area]

    xc = data.xcp.values[area]
    yc = data.ycp.values[area]

    c = data[tracer].values[area]
    c[~isfinite] = np.nan

    if get_wind:
        if 'wind_eff_%s' % tracer[:3] in data:
            wind = data['wind_eff_%s' % tracer[:3]].values[area]
        else:
            ueff = data['u_eff_%s' % tracer[:3]].values[area]
            veff = data['u_eff_%s' % tracer[:3]].values[area]
            wind = np.sqrt(ueff**2 + veff**2)

        wind[~isfinite] = np.nan
    else:
        wind = None


    p = data['detected_plume'].values[area]

    # estimate noise
    noise_level = data[tracer].attrs.get('noise_level', np.nan)

    if isinstance(noise_level, str):
        molar_mass = 46.0055e-3 # of NO2
        noise_level = {
            'low':  1e15 * molar_mass / N_A * 1e4, # molec cm-2 -> kg m-2
            'high': 2e15 * molar_mass / N_A * 1e4,
            'S5':   1.3e15 * molar_mass / N_A * 1e4
        }[noise_level]

    c_std = np.full(np.shape(c), noise_level)
    s = None

    if dy is not None:
        s = np.arange(ya + 0.5 * dy, yb, dy)

        sub_areas = [(y0 - 0.5 * dy <= yy) & (yy < y0 + 0.5 * dy)
                     for y0 in s]

        xx = get_values_from_areas(xx, sub_areas)
        yy = get_values_from_areas(yy, sub_areas)

        xc = get_values_from_areas(xc, sub_areas)
        yc = get_values_from_areas(yc, sub_areas)

        c = get_values_from_areas(c, sub_areas)
        c_std = get_values_from_areas(c_std, sub_areas)

        p = np.array([np.sum(v, dtype=int) for v in
                      get_values_from_areas(p, sub_areas)])

        if get_wind:
            wind = np.nanmean(wind) if np.any(np.isfinite(wind)) else np.nan

    return s, xx, yy, xc, yc, c,  c_std, p, wind




def get_values_from_areas(values, sub_areas):
    return np.array([values[a] if np.any(a) else np.array([]) for a in sub_areas])



def extract_true_line_density(data, tracer, xa, xb, ya, yb):


    # only use pixels that are valid observations
    xp = data['xp'].values
    yp = data['yp'].values

    # use area of true plume if larger than [ya,yb]
    area_true = (xa <= xp) & (xp < xb) & data['%s_true_plume' % tracer[:3]].values

    if np.any(area_true):
        #ya_true, yb_true = get_plume_width(data, dy=5e3, area=area_true)
        ya_true, yb_true = -40e3, 40e3
        ya = min(ya, ya_true)
        yb = max(yb, yb_true)

    # area
    area = (xa <= xp) & (xp < xb) & (ya <= yp) & (yp <= yb)

    # get wind field
    if 'wind_eff_%s' % tracer[:3] in data:
        wind = data['wind_eff_%s' % tracer[:3]].values[area]
    else:
        ueff = data['u_eff_%s' % tracer[:3]].values[area]
        veff = data['v_eff_%s' % tracer[:3]].values[area]
        wind = np.sqrt(ueff**2 + veff**2)

    # get data
    tracer = '%s_true_signal_mass' % tracer[:3]
    c = data[tracer].values[area]

    # include wind speeds at pixel using weights
    c = (c * wind) / np.nanmean(wind)

    line_density = np.sum(c * 2e3**2 / (xb -xa))


    # wind speed
    if 'wind_eff_%s' % tracer[:3] in data:
        wind = data['wind_eff_%s' % tracer[:3]].values[area]
        wind = np.nanmean(wind)

    else:
        ueff = data['u_eff_%s' % tracer[:3]].values[area]
        veff = data['v_eff_%s' % tracer[:3]].values[area]
        wind = np.nanmean(np.sqrt(ueff**2 + veff**2))

    return line_density, yp[area], c, wind


def get_plume_width(data, dy=5e3, area='detected_plume'):

    # distance from center line
    if isinstance(area, str):
        yp = data['yp'].values[data[area]]
    else:
        yp = data['yp'].values[area]

    ymin = np.floor((yp.min() - 2 * dy) / dy) * dy
    ymax = np.ceil((yp.max() + 2 * dy) / dy) * dy

    return ymin, ymax


def extract_line_density(data, tracer, xa, xb, ya=None, yb=None, dy=5e3,
                         method='both', show=False, show_true=True,
                         add_errors=False):
    """
    Extract line densitiy of tracer in data within polygon given by along-plume
    interval [xa, xb] and across-plume interval [ya, yb]. If ya or yb is None
    estimate plume width from detected pixels. The plume center line is
    described by the `curve`.

    The function uses two methods:
        - fitting a Gaussian curve
        - summing mean values in sub-areas in across-plume direction using
          sub-areas with width `dy`
    """
    if ya is None or yb is None:
        ya, yb = get_plume_width(data)

    s = np.linspace(ya, yb, 501)


    # true line density
    true_density, y_true, c_true, wind_true = extract_true_line_density(data, tracer, xa, xb, ya, yb)

    # returns
    results = {
        'gauss': np.nan,
        'gauss_std': np.nan,
        'sub-areas': np.nan,
        'sub-areas_std': np.nan,
        'area': (
            (xa <= data.xp) & (data.xp < xb) &
            (ya <= data.yp) & (data.yp <= yb) &
            data['%s_isfinite' % tracer[:3]].values
        ),
        'density_true': true_density,
        'wind_true':    wind_true
    }

    if method == 'gauss' or method == 'both':

        # extract pixels in area
        _, x, y, xc, yc, c, c_std, detection, wind = extract_pixels(data, tracer, xa, xb,
                                                                    ya, yb,
                                                                    dy=None,
                                                                    get_wind=True)

        # fit function (Gaussian or Error function)
        mass_gauss, mass_gauss_std, fit_gauss = fit_function(y, c, c_std, s, function='gauss')

        if mass_gauss < 0.0 or mass_gauss > 1000.0 or mass_gauss_std > 1000.0:
            print('Fit failed:', mass_gauss, mass_gauss_std)
            results['gauss'] = np.nan
            results['gauss_std'] = np.nan
        else:
            results['gauss'] = mass_gauss
            results['gauss_std'] = mass_gauss_std

        if wind is not None:
            results['wind'] = np.nanmean(wind)

    if method == 'sub-areas' or method == 'both':

        # extract pixels in sub-areas
        ss, xx, yy, _, _, cc, cc_std, n_detection, wind = extract_pixels(data, tracer, xa, xb, ya, yb,
                                                                         dy=dy, get_wind=True)

        # sum mean values of sub-areas (interp missing values)
        area_means = np.array([np.nanmean(c) if np.any(np.isfinite(c))
                               else np.nan for c in cc])

        # TODO: error of interpolation
        area_means_std = np.array([np.nanmean(c) / np.sum(np.isfinite(c)) if
                                   np.any(np.isfinite(c)) else np.nan for c in cc_std])


        valids = np.isfinite(area_means)
        bad_lines = np.isnan(area_means) & (n_detection > 0)

        if np.all(~valids) | np.any(bad_lines):
            area_means[:] = np.nan
        else:
            area_means = np.interp(ss, ss[valids], area_means[valids],
                                   left=0.0, right=0.0)

        mass_means = np.sum(area_means * dy)
        mass_means_std = np.nansum(area_means_std * dy)

        results['sub-areas'] = mass_means
        results['sub-areas_std'] = mass_means_std

        if wind is not None:
            if np.any(np.isfinite(wind)):
                results['wind'] = np.nanmean(wind)
            else:
                results['wind'] = np.nan


    # set gauss to nan if missing values in sub-areas
    if np.isnan(results['sub-areas']):
        results['gauss'] = np.nan


    if show:
        print('true wind: %f' % wind_true)
        fmt = {
            'CO2': '%.0f$\pm$%.0f$\,$kg$\,$m$^{-1}$',
            'NO2': '%.0f$\pm$%.0f$\,$g$\,$m$^{-1}$',
        }[tracer[:3]]

        if tracer.startswith('NO2'):
            fmt = '%.0f$\pm$%.0f$\,$g$\,$m$^{-1}$'
            true_fmt = '%.0f$\,$g$\,$m$^{-1}$'
            factor = 1e6

        else: # CO2
            fmt = '%.0f$\pm$%.0f$\,$kg$\,$m$^{-1}$'
            true_fmt = '%.0f$\,$kg$\,$m$^{-1}$'
            factor = 1e3

        # third of page
        figsize = np.array(plt.rcParams['figure.figsize']) * 2 / 3
        fig = plt.figure(figsize=figsize)
        ax = fig.add_axes([0.16,0.17,0.82,0.8])

        if show_true:
            ax.plot(y_true/1e3, factor*c_true, 'D', color='tab:blue', ms=3,
                    label='Model tracer (%s)' % true_fmt % true_density)

        if not add_errors:
            ax.plot(y[detection]/1e3, factor*c[detection],
                        ls='', marker='o', markersize=5, color='tab:red',
                        label='Inside detected plume')

            ax.plot(y[~detection]/1e3, factor*c[~detection],
                        ls='', marker='o', markersize=5, color='tab:red', mfc='none',
                        label='Outside detected plume')

        else:
            ax.errorbar(y[detection]/1e3, factor*c[detection], factor*c_std[detection],
                        ls='', marker='o', markersize=5, color='tab:red',
                        label='Inside detected plume')

            ax.errorbar(y[~detection]/1e3, facotr*c[~detection], factor*c_std[~detection],
                        ls='', marker='o', markersize=5, color='tab:red', mfc='none',
                        label='Outside detected plume')

        ax.plot(ss/1e3, factor*area_means, lw=2, ls='steps-mid', color='black',
                label='Sub-area means (%s)' % fmt % (factor*mass_means/1e3,
                                                     factor*mass_means_std/1e3))

        ax.plot(s/1e3, factor*fit_gauss, lw=2, color='black',
                label='Gaussian function (%s)' % fmt %
                (factor*mass_gauss/1e3, factor*mass_gauss_std/1e3), ls='--')


        ax.set_ylim(bottom=min(0.0, ax.get_ylim()[0]))

        ax.grid(True)
        ax.legend(fontsize='small', ncol=1)
        ax.set_xlabel('Across-plume direction (km)')

        if tracer.startswith('NO2'):
            ax.set_ylabel('%s$_2$ columns (mg m$^{-2}$)' % tracer[:2])
        else:
            ax.set_ylabel('%s$_2$ columns (g m$^{-2}$)' % tracer[:2])



    return results




def extract_along_polygon(data, tracer, xa, xb, ya, yb, dy=5e3, curve=None, show=False,
                         function=None):
    """
    Extract all pixels between on the interval [a,b[ using the pixel centers.
    """
    s = np.linspace(ya, yb, 501)

    if tracer.startswith('both'):
        tracers = ['CO2'+tracer[4:], 'NO2'+tracer[4:]]
    else:
        tracers = [tracer]


    # number of expected pixels in area
    n_expected = (xb - xa) * (yb - ya) / 2000**2

    # mask for area of interest
    x = np.array(data.xp)
    y = np.array(data.yp)
    area = (xa <= x) & (x < xb) & (ya <= y) & (y <= yb)

    lines = []


    # fit CO2 and NO2 plume simultaneously

    if len(tracers) == 2 and function is not None:
        x1, y1, xc1, yc1, c1, c_std1 = extract_pixels(data, tracers[0], area)
        x2, y2, xc2, yc2, c2, c_std2 = extract_pixels(data, tracers[1], area)

        co2_mass, no2_mass, co2_fit, no2_fit  = fit_two_gaussian(y1, c1,
                                                                 c_std1, y2,
                                                                 c2, c_std2, s=s)

    for i, tracer in enumerate(tracers):

        x, y, xc, yc, c, c_std, pixels = extract_pixels(data, tracer, area)

        n_pixels = np.sum(c.size)
        n_valids = np.sum(np.isfinite(c))


        # fit gaussian or error function
        if function is not None:
            mass_fit, mass_fit_std,  gauss_curve = fit_function(y, c, c_std, s=s,
                                                                function=function)

        if show:
            colors = ['#1f77b4', '#ff7f0e']

            if i == 0:
                fig = plt.figure()
                ax = fig.add_subplot(111)
            else:
                ax = ax.twinx()

            lines += ax.plot(y / 1e3, c, marker='os'[i],
                             color=colors[i], ms=4, ls='',
                             label='CO2 mass')

            totals = 0.0
            for yy in np.arange(ya, yb, dy):
                sub_area = (yy <= y) & (y < yy + dy)

                ax.plot(y[sub_area] / 1e3, c[sub_area], '.')

                totals += np.mean(c[sub_area]) * 10e3 * 5e3

            totals /= 10e3


            # fitted gaussian function
            if function is not None:
                lines += ax.plot(s/1e3, gauss_curve, label='%s (%.3f kg m$^{-1}$)' %
                                 (function, mass_fit), color=colors[i])

            # plot double fit
            if len(tracers) == 2 and function is not None:
                lines += ax.plot(s/1e3, [co2_fit, no2_fit][i],
                                 label='Gaussian (%.1f kg m$^{-1}$)' % [co2_mass, no2_mass][i],
                                 color='rg'[i], ls='--')


            ax.set_xlabel('across-plume direction (km)')
            ax.set_xlim(ya/1e3, yb/1e3)

            if tracer.startswith('CO2'):
                ax.set_ylabel('CO$_2$ mass (kg m$^{-2}$)')
            else:
                ax.set_ylabel('NO$_2$ mass (kg m$^{-2}$)')


        # wind
        if 'u_10m' in data:
            u10 = np.array(data['u_10m'])[pixels]
            v10 = np.array(data['v_10m'])[pixels]

        if 'u_eff' in data:
            u = np.array(data['u_eff'])[pixels]
            v = np.array(data['v_eff'])[pixels]

        # compute normal wind speed at pixel
        if curve is not None:
            t0 = curve.get_parameter(data.x.values[pixels], data.y.values[pixels])
            winds = []

            for i,(uu,vv) in enumerate(zip(u, v)):
                wind = np.dot([uu,vv], curve.compute_tangent(t0=t0[i], norm=True))
                winds.append(wind)

            winds = np.array(winds)


        if n_valids < 0.5 * n_expected:
            mass = np.nan
            mass_std = np.nan
            flux = np.nan

        else:
            # pixels
            polygons = [Polygon(np.transpose(x)) for x in zip(xc,yc)]

            # areas
            areas = np.array([px.area for px in polygons])

            # integral in kg/m
            mass = np.nansum(c * areas) / (xb - xa)

            # estimate noise
            noise_level = data[tracer].attrs.get('noise_level', np.nan)

            if isinstance(noise_level, str):
                molar_mass = 46.0055e-3 # of NO2
                mass_std = {
                    'low':  1e15 * molar_mass / N_A * 1e4, # molec cm-2 -> kg m-2
                    'high': 2e15 * molar_mass / N_A * 1e4,
                    'S5':   1.3e15 * molar_mass / N_A * 1e4
                }[noise_level]
            else:
                mass_std = np.sum(noise_level * areas) / (xb - xa) / np.sqrt(np.sum(pixels))


            # flux using true wind speed and direction
            flux = np.sum(winds * c * areas) / (xb - xa)

    if show:
        plt.legend(lines, [l.get_label() for l in lines])
        plt.tight_layout()


    return mass, mass_std, flux, pixels





