

from sympy.solvers import solve
import sympy

import matplotlib.pyplot as plt
import numpy as np
import scipy
import xarray


def solve_min_distance_to_curve():
    """
    Find coefficients of cost function for quadratic polynomial curve.
    """
    t = sympy.Symbol('t')

    qx = sympy.Symbol('qx')
    qy = sympy.Symbol('qy')

    a = [sympy.Symbol('a[%d]' % i) for i in range(3)]
    b = [sympy.Symbol('b[%d]' % i) for i in range(3)]

    px = a[0] * t**2 + a[1] * t + a[2]
    py = b[0] * t**2 + b[1] * t + b[2]


    # cost function: minimize difference
    f = (px - qx)**2 + (py - qy)**2

    d = f.diff(t).expand().collect(t)

    c0 = d.coeff(t, 3)
    c1 = d.coeff(t, 2)
    c2 = d.coeff(t, 1)
    c3 = d.coeff(t, 0)

    print(c0)
    print(c1)
    print(c2)
    print(c3)


def solve_arc_length():
    t = sympy.Symbol('t')

    t0 = sympy.Symbol('t0')
    t1 = sympy.Symbol('t1')

    a = [sympy.Symbol('a[%d]' % i) for i in range(3)]
    b = [sympy.Symbol('b[%d]' % i) for i in range(3)]

    px = a[0] * t**2 + a[1] * t + a[2]
    py = b[0] * t**2 + b[1] * t + b[2]


    f = px.diff(t)**2 + py.diff(t)**2
    f = f.expand().collect(t)

    c1 = f.coeff(t, 2)
    c2 = f.coeff(t, 1)
    c3 = f.coeff(t, 0)

    print(c1)
    print(c2)
    print(c3)



def integral_sqrt_poly(x, a, b, c):
    """
    Integral over sqrt(a*x**2 + b*x + c)
    """
    s = np.sqrt( a * x**2 + b * x + c)

    A = (b + 2 * a * x) / (4 * a) * s
    B = (4 * a * c - b**2) / (8 * a**(3/2))
    C = np.abs(2 * a * x + b + 2 * np.sqrt(a) * s)

    return A + B * np.log(C)


def compute_arc_length(curve, smin, smax):
    a, b = curve.get_coefficients()

    c0 = 4 * a[0]**2 + 4 * b[0]**2
    c1 = 4 * a[0] *a[1] + 4 * b[0] * b[1]
    c2 = a[1]**2 + b[1]**2

    smin = integral_sqrt_poly(smin, c0, c1, c2)
    smax = integral_sqrt_poly(smax, c0, c1, c2)

    return smax - smin

compute_arc_length = scipy.vectorize(compute_arc_length)


def quadratic_equation(a, b, c):
    """
    Find roots of:
        a * x**2 + b * x + c
    """
    p = np.sqrt(b**2 - 4.0 * a * c)

    x0 = (-b + p) / (2 * a)
    x1 = (-b - p) / (2 * a)

    return np.array([x0, x1])


def cubic_equation(a,b,c,d):
    """
    Find roots of cubic polynomial:
        a * x**3 + b * x**2 + c * x + d = 0
    """
    dtype = np.complex256
    a = np.asarray(a).astype(dtype)
    b = np.asarray(b).astype(dtype)
    c = np.asarray(c).astype(dtype)
    d = np.asarray(d).astype(dtype)

    d0 = b**2 - 3 * a * c
    d1 = 2 * b**3 - 9 * a * b * c + 27 * a**2 * d

    C = ((d1 + np.sqrt(d1**2 - 4 * d0**3)) / 2.0)**(1/3)

    xi = (-1.0 + np.sqrt(-3.0 + 0j)) / 2.0
    s = lambda k: xi**k * C

    roots = [-1.0/ (3.0 * a) * (b + s(k) + d0 / s(k)) for k in range(3)]

    return np.array(roots)



def arc_length_of_2d_curve(curve, a, b):

    t = np.linspace(a,b, 5001)
    xt, yt = curve(t=t, m=1)

    v = np.sqrt( xt**2 + yt**2)

    l = scipy.integrate.simps(v, t)

    return l

arc_length_of_2d_curve = scipy.vectorize(arc_length_of_2d_curve)




def distance_from_curve(t, curve, qx, qy):
    """get distance from curve"""
    px, py = curve(t=t)
    return np.sqrt((px - qx)**2 + (py - qy)**2)


def compute_plume_coordinates(data, curve, show=False,
                              which='centers'):
    """
    Computes along- and across-plume coordinates analytically
    if curve.degree == 2.

    Parameters
    ----------
    data : satellite data incl. x, y and plume_area
    curve : center curve
    which : process either pixel 'centers' or 'corners'.

    """
    if curve.degree != 2:
        raise ValueError('Degree of curve needs to be 2 not %d' % curve.degree)

    a, b = curve.get_coefficients()

    # parameter for minimum distance to curve
    if 'plume_area' in data:
        area = data.plume_area.values
    else:
        area = np.ones(data.x.shape, bool)


    if which == 'centers':
        qx = data.x.values[area]
        qy = data.y.values[area]
    else:
        qx = data.xc.values[area].flatten()
        qy = data.yc.values[area].flatten()

    c0 = 4*a[0]**2 + 4*b[0]**2
    c1 = 6*a[0]*a[1] + 6*b[0]*b[1]
    c2 = 4*a[0]*a[2] - 4*a[0]*qx + 2*a[1]**2 + 4*b[0]*b[2] - 4*b[0]*qy + 2*b[1]**2
    c3 = 2*a[1]*a[2] - 2*a[1]*qx + 2*b[1]*b[2] - 2*b[1]*qy


    if False:#np.abs(c0) < 1e-9:
        roots = quadratic_equation(c1, c2, c3)
        roots.astype(np.complex_)
    else:
        roots = cubic_equation(c0, c1, c2, c3)


    real = np.abs(roots.imag) < 1e-6

    tmin = []
    n_no_solutions = 0
    n_multiple_solutions = 0


    for i in range(qx.size):

        n_solutions = np.sum(real[:,i])

        if n_solutions == 0:
            tmin.append(np.nan)
            n_no_solutions += 1

        elif n_solutions == 1:
            tmin.append( float(roots[:,i][real[:,i]].real) )

        elif n_solutions > 1:
            j = np.argmin(
                distance_from_curve(roots[:,i].real, curve, qx[i], qy[i])
            )
            tmin.append(roots[j,i].real)
            n_multiple_solutions += 1

        else:
            raise ValueError


    if n_no_solutions > 0:
        name = ' '.join(str(v) for v in [data.time, data.orbit, data.lon_eq])
        print('No real solution for some points in "%s"' % name)
        #raise ValueError('No real solution for some points in "%s"' %
        #                 name)

    if n_multiple_solutions > 0:
        name = ' '.join(str(v) for v in [data.time, data.orbit, data.lon_eq])
        print('%d multiple solutions in %s' % (n_multiple_solutions, name))


    tmin = np.array(tmin)
    px, py = curve(t=tmin)

    # sign of distance (negative left of curve from source)
    t = curve.get_parameter(qx, qy)
    v = curve.compute_tangent(t, norm=True)
    n = np.array([px-qx, py-qy])
    cross = np.cross(v, n, axis=0)
    sign = np.sign(cross)

    if which == 'centers':

        # compute distance
        distance = xarray.full_like(data.x, np.nan)
        distance.values[area] = sign * np.sqrt((px - qx)**2 + (py - qy)**2)

        # arc-length
        arc = xarray.full_like(data.x, np.nan)
        arc.values[area] = compute_arc_length(curve, curve.t_o, tmin)

    else:
        # distance
        d = sign * np.sqrt((px - qx)**2 + (py - qy)**2)
        distance = xarray.full_like(data.xc, np.nan)
        distance.values[area] = d.reshape(d.size//4, 4)

        # arc-length
        a = compute_arc_length(curve, curve.t_o, tmin)
        arc = xarray.full_like(data.xc, np.nan)
        arc.values[area] = a.reshape(a.size//4, 4)

    if show:
        fig = plt.figure()
        ax = plt.subplot(aspect='equal')

        for i, (x,y) in enumerate(zip(qx, qy)):
            px, py = curve(t=tmin[i])
            ax.plot([px, x], [py, y], 'o-')

        ax.plot(*curve(t=np.linspace(0, tmin.max())), 'k-')


    return arc, distance



