
from collections import defaultdict
import itertools
import functools
import operator

from matplotlib.dates import DateFormatter
import matplotlib.pyplot as plt
import numpy as np
import pandas
import scipy.stats
import scipy.signal
import statsmodels.api as sm

import smartcarb.misc
from amrs.misc.chem import ppm_to_mass, kgs_to_Mtyr

import emiss.gauss
from . import timeprofiles as tp
from . import io


COLORS = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd',
          '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']

LON_EQ_TO_LETTER = {
    0: 'a',  401: 'b', 805: 'c', 1207: 'd', 1610: 'e', 2012: 'f'
}

def extract_time_range(time, values, tmin, tmax):

    data = []

    for hour in range(tmin, tmax+1):
        mask = np.array([pandas.Timestamp(t).hour == hour for t in
                         np.array(time)])

        data.append( values[mask] )

        new_time = np.array(time)[mask]


    return new_time, np.array(data)




def init_mass_balance_plot(tracer, distances, delta, emissions, q_true, u_1km):

    xvalues = (distances[1:] - delta / 2) / 1e3 # m -> km

    fig = plt.figure()

    # [left, bottom, width, height]
    ax =  fig.add_axes([0.12, 0.39, 0.85, 0.57])
    ax2 = fig.add_axes([0.12, 0.12, 0.85, 0.23])

    lines = []


    lines.append(
        ax.fill_betweenx([-20,80], [-10,-10], [+10,+10], color='#eeeeee',
                          label='City area')
    )
    lines.append(
        ax2.fill_betweenx([0,80], [-10,-10], [+10,+10], color='#eeeeee',
                          label='_none_')
    )

    lines += ax.plot([-50, 200],
                     [q_true, q_true], 'k-', label='True emissions (10:30 UTC)')

    label = 'Range of true emissions (6-12 UTC)'
    lines.append(
        ax.fill_between([-50, 200],
                        [min(emissions), min(emissions)],
                        [max(emissions), max(emissions)],
                        label=label, color='silver')
    )

    # constant wind speed
    lines += ax2.plot(xvalues, np.full(xvalues.shape, u_1km), 's-',
                       label='wind speed over Berlin (0 - 500 m)')

    ax2.set_xlabel('Along plume distance from source (km)')

    if tracer == 'CO2':
        ax.set_ylabel('CO$_2$ flux (Mt CO$_2$ yr$^{-1}$)')
    else:
        ax.set_ylabel('NO$_x$ flux (kt NO$_2$ yr$^{-1}$)')

    return ax, ax2, lines



def add_mass_balance_plot(i, ax, ax2, tracer, distances, delta, flux, flux_std,
                          ueff, lines=None, add_wind=False, add_result=True,
                          dq=np.nan, fit=None):

    if lines is None:
        lines = []

    xvalues = (distances[1:] - delta / 2) / 1e3 # m -> km

    if add_wind:
        lines += ax2.plot(xvalues, ueff, 'o-',
                     label='wind speed in plume')

    factor = 1.0 if  tracer == 'CO2' else 1e3

    label = [
        'Model tracer + winds',
        'Noise-free model tracer',
        'Wind error',
        'Instrument error',
        'Background error',
        'Total error',
    ][i]
    marker = ['D', 'o', 's', 'd', 'v', '^'][i]

    #extra = '$\Delta$Q = %+.1f %st yr$^{-1}$' % (dq, 'k' if tracer ==
    #                                     'NO2' else 'M')
    #label = labels[i] % extra

    if add_result:
        if np.any(np.isnan(flux_std)):
            lines += ax.errorbar(xvalues, factor * kgs_to_Mtyr(flux),
                                 ls='', marker=marker,
                                 label=label, color=COLORS[i])

        else:
            lines.append( ax.errorbar(xvalues, factor * kgs_to_Mtyr(flux),
                                 factor * kgs_to_Mtyr(flux_std),
                                 ls='', marker=marker,
                                 label=label, color=COLORS[i])
                        )

        ax.plot(xvalues, fit, label='__none__',
                color=COLORS[i])

    ax.set_xticklabels([])

    ax.set_xlim(min(distances)/1e3, max(distances)/1e3)
    ax2.set_xlim(min(distances)/1e3, max(distances)/1e3)

    return lines



def finalize_mass_balance_plot(ax, ax2, tracer, lines):

    _, ymax = ax.get_ylim()
    ax.set_ylim(bottom=-20, top=max(ymax, 80 if tracer == 'NO2' else 80))

    ax2.set_ylabel('Wind speed (m s$^{-1}$)', labelpad=13)
    ax2.set_ylim(0, 20)

    ax.set_zorder(ax2.get_zorder()+1)
    ax.patch.set_visible(False)

    ax.legend(loc='upper left', fontsize='small', ncol=2)
    ax2.legend(loc='upper left', fontsize='small')

    ax.grid(True)
    ax2.grid(True)

    #plt.gcf().align_ylabels()


def compute_bias_and_scatter(ts, tracer, n_knots):
    bias = []
    scatter = []
    for i in [1,2,3,4,5]:
        diffs, mb, sd = plot_annual_emissions_by_constellation(ts,
                                                               n_knots=n_knots,
                                                               tracer=tracer,
                                                               constellation=i,
                                                               show=False,
                                                               gamma=None)
        bias.append(mb)
        scatter.append(sd)

    return np.array(bias), np.array(scatter)


def plot_bias_and_scatter_for_number_of_satellites(ts_analy, ts_mb,
                                                   tracer='CO2', n_knots=3):
    """
    """
    bias_anal, scatter_anal = compute_bias_and_scatter(ts_analy, tracer,
                                                       n_knots)
    bias_mb, scatter_mb = compute_bias_and_scatter(ts_mb, tracer, n_knots)

    fig, axes = plt.subplots(1, 2)
    nsats = np.arange(1,6)

    axes[0].plot(nsats, 100 * np.array(bias_anal), 'o--',
                 label='analytical inversion')

    axes[0].plot(nsats, 100 * np.array(bias_mb), 's--',
                 label='mass-balance')

    axes[1].plot(nsats, 100 * np.array(scatter_anal), 'o--',
                 label='analytical inversion')

    axes[1].plot(nsats, 100 * np.array(scatter_mb), 's--',
                 label='mass-balance')

    axes[0].set_ylabel('bias (%)')
    axes[1].set_ylabel('scatter (%)')

    for i in [0,1]:
        axes[i].grid(True)
        axes[i].set_xlabel('number of satellites')
        axes[i].legend()
        axes[i].set_xlim(0.5, 5.5)
        axes[i].set_ylim(0, 35)

        axes[i].set_xticks(nsats)

    plt.tight_layout()

    return bias_anal, scatter_anal, bias_mb, scatter_mb




def plot_annual_emissions_by_constellation(data, n_knots=3, tracer='CO2',
                                           constellation='eq', show=True,
                                           gamma=None, is_constant=False):

    # constellations
    if constellation == 'eq': # equidistant spacing
        nsats = ['a', 'b', 'c', 'd', 'e', 'f', # TODO:use 'abc' etc.
                 'ad', 'be', 'cf', 'ace', 'bdf', 'abcdef']

        steps = [(1,0,6), (2,6,9), (3,9,11), (6,11,12)]
    else:
        nsats = [''.join(c) for c in itertools.combinations('abcdef', constellation)]
        steps = [(constellation,0,len(nsats))]

    annual_emissions = []
    annual_emissions_std = []
    overpasses = []

    for nsat in nsats:
        cycle = tp.add_cycle(data, n_knots=n_knots, nsat=nsat, gamma=gamma)
        em = cycle['fitted_cycle'].attrs['annual_mean']
        em_std = cycle['fitted_cycle'].attrs['annual_mean_std']
        annual_emissions.append(em)
        annual_emissions_std.append(em_std)

        overpasses.append(
            cycle['fitted_cycle'].attrs['number of overpasses']
        )

    annual_emissions = np.array(annual_emissions)
    annual_emissions_std = np.array(annual_emissions_std)
    overpasses = np.array(overpasses)

    # true emissions
    true_emissions = {'CO2': 20.0, 'NO2': 22.1}[tracer]


    #
    mask = annual_emissions_std > true_emissions
    if np.any(mask):
        print('Est uncertainty larger than true emissions')
        annual_emissions[mask] = np.nan
        annual_emissions_std[mask] = np.nan


    differences = []
    bias = []
    scatter = []

    if show:
        plt.figure()

    for nsat, i, j in steps:

        if show:

            em_std = r'$\bar{\sigma}$ = %.1f Mt yr$^{-1}$' % np.nanmean(annual_emissions_std[i:j])

            if nsat == 1:
                label = '1 satellite (%s)' % em_std
            else:
                label = '%d satellites (%s)' % (nsat, em_std)

            plt.bar(np.arange(i,j), annual_emissions[i:j], tick_label=nsats[i:j],
                    label=label)

            plt.errorbar(np.arange(i,j), annual_emissions[i:j], annual_emissions_std[i:j],
                         ls='', marker='', color='k', capsize=3)

            nans = np.isnan(annual_emissions[i:j])
            plt.plot(np.arange(i,j)[nans], np.full(nans.sum(), 2.5), 'kx', ms=15)

        difference = annual_emissions[i:j] - true_emissions
        mb = np.nanmean(difference) / true_emissions
        sd = np.nanstd(difference - true_emissions) / true_emissions

        differences.append(difference)
        bias.append(mb)
        scatter.append(sd)


    if show:
        if tracer == 'CO2':
            if is_constant:
                plt.plot([-0.5, len(nsats)-0.5], [16.8, 16.8], 'k--',
                         label='Emissions at overpass')
            else:
                plt.plot([-0.5, len(nsats)-0.5], [20.0, 20.0], 'k--',
                         label='Emissions at overpass')
            plt.ylabel('CO$_2$ emissions (Mt yr$^{-1}$)')
        else:
            plt.plot([-0.5, 11.5], [22.1, 22.1], 'k-',
                     label='Emissions at overpass')
            plt.ylabel('NO$_x$ emissions (kt yr$^{-1}$)')

        plt.xlabel('Constellations', labelpad=-10)

        plt.xticks(np.arange(len(nsats)), nsats, rotation=45)

        plt.legend(ncol=2, fontsize='small')

        plt.ylim(0, 50)

        for i, o in enumerate(overpasses):
            if np.isnan(annual_emissions[i]):
                y = 5.0
            elif np.isnan(annual_emissions_std[i]):
                y = annual_emissions[i]
            else:
                y = annual_emissions[i] + annual_emissions_std[i]

            plt.text(i, y + 0.8,
                     '(%s)' % o, fontsize='small',
                     ha='center')



        plt.tight_layout()

    print(annual_emissions_std)

    return np.array(differences), np.array(mb), np.array(sd)



def plot_ts(tracer, data, cycle=None, nsat=1, label=None, origin='Berlin',
            n_knots=3, gamma=None, use_cycle_error=False, show_errors=True):

    # plot results
    if tracer == 'NO2':
        units = 'kt yr$^{-1}$'
    elif tracer == 'CO2':
        units = 'Mt yr$^{-1}$'
    else:
        units = 'ratio'


    if tracer == 'ratio':
        var_true = 'true_ratio'
        var_est = 'est_ratio'
        var_std = 'est_ratio_std'
    else:
        var_true = 'true_emissions'
        var_est = 'est_emissions'
        var_std = 'est_emissions_std'

    fig = plt.figure(figsize=(8.4, 3.5))
    ax = fig.add_subplot(111)


    em_date, em_range = extract_time_range(data['time'],
                                           data[var_true], 10, 11)


    ax.plot(em_date, em_range.mean(0), 'k-',
            label='Emissions at overpass (10-11 UTC: %.1f %s)' % (em_range.mean(),
                                                           units))

    ax.fill_between(em_date, em_range.min(0), em_range.max(0), color='silver')


    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)

    # statistics
    valids = np.isin(data.lon_eq, lon_eqs_in_constellation)
    true_values = data[var_true].loc[data.overpass].values[valids]
    est_values = data[var_est].values[valids]

    mb = np.nanmean(est_values - true_values)
    sd = np.nanstd(est_values - true_values)
    rmsd = np.sqrt(mb**2 + sd**2 / sd.size)

    print(mb, sd, rmsd)


    # fit seasonal cycle
    tp.add_cycle(data, n_knots=n_knots, nsat=nsat, gamma=gamma)


    for i, lon_eq in enumerate(lon_eqs_in_constellation, 1):

        letter = LON_EQ_TO_LETTER[lon_eq]

        sat = data['lon_eq'] == lon_eq

        times = data['overpass'].loc[sat].values
        values = data[var_est].loc[sat].values

        # use or not use uncertainty estimated from fitted cycle
        if use_cycle_error:
            errors = np.sqrt(data['fitted_cycle'].attrs['chi2'])
        else:
            errors = data[var_std].loc[sat].values

        n_sats = np.sum(np.isfinite(values))

        if show_errors:
            ax.errorbar(times, values, errors, ls='', marker='o',
                        label='#%s (%d plumes)' % (letter, n_sats))
        else:
            ax.errorbar(times, values, ls='', marker='o',
                        label='#%s (%d plumes)' % (letter, n_sats))

    if label is not None:
        for i,(t,v) in enumerate(zip(times,values)):
            s = '%.2g' % ts[label][i]
            ax.text(t,v,s)

    ax.xaxis.set_major_formatter(DateFormatter('%b'))

    ax.set_xlim('2015-01-01', '2015-12-31')
    if origin == 'Berlin':
        if tracer == 'ratio':
            ax.set_ylim(200,1800)
        else:
            ax.set_ylim(-10, max(60, np.nanmax(data[var_est]) ))
    else:
        ax.set_ylim(ymin=0)

    ax.set_xlabel('2015')

    if tracer == 'CO2':
        ax.set_ylabel('CO$_2$ emissions (%s)' % units)
    elif tracer == 'NO2':
        ax.set_ylabel('NO$_2$ emissions (%s)' % units)
    else:
        ax.set_ylabel('CO$_2$-to-NO$_2$ emission ratio')

    # plot cycle
    if tracer != 'ratio' and 'fitted_cycle' in data:
        ax.plot(data['time'], data['fitted_cycle'], 'r-',
                label='Fitted seasonal cycle (%.1f$\pm$%.1f %s)' %
                (data['fitted_cycle'].attrs['annual_mean'],
                 data['fitted_cycle'].attrs['annual_mean_std'],
                 units))

    if tracer == 'ratio' and 'fitted_ratio_cycle' in data:
        ax.plot(data['time'], data['fitted_ratio_cycle'], 'r-',
                label='Fitted seasonal cycle (%.1f$\pm$%.1f %s)' %
                (data['fitted_ratio_cycle'].attrs['annual_mean'],
                 data['fitted_ratio_cycle'].attrs['annual_mean_std'],
                 units))


    ax.grid(True, ls='-', color='gray')

    if isinstance(nsat, str):
        nsat = int(nsat[:1])

    ncols = {
        1: 1,
        2: 3,
        3: 3,
        4: 3,
        5: 3,
        6: 3}[nsat]

    plt.legend(loc=1, ncol=ncols, fontsize='small')
    plt.tight_layout(pad=0.25)
    fig.subplots_adjust(top=0.95)



def filter_nsat(d, nsat):
    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)
    valids = np.isin(d.lon_eq, lon_eqs_in_constellation)
    return d.loc[{'overpass': d.overpass.values[valids]}]


def plot_mb_histograms(method='sub-areas', co2_ssp='medium',
                       no2_ssp='high', overpass=None, show=True):
    """
    Plot histograms showing deviation between estimated and true emissions
    of the mass-balance approach
    """

    ts = {}
    mean_bias = {}
    scatter = {}
    numbers = {}

    for i, var in enumerate(['method', 'retrieval', 'background', 'wind',
                             'combined']):

        ts[var] = io.read_mass_balance('CO2', which=var, method=method,
                                       co2_ssp=co2_ssp, no2_ssp=no2_ssp)


    overpass = [v.overpass.values[np.isfinite(v.est_emissions.values)] for v in ts.values()]
    overpass = [set(o) for o in overpass]
    overpass = functools.reduce(operator.and_, overpass)
    overpass = np.array(list(overpass))


    for key, values in ts.items():
        ts[key] = values.loc[{"overpass": overpass}]


    if show:
        fig, axes = plt.subplots(ncols=5, figsize=(14, 3.0))

    for i, var in enumerate(['method', 'retrieval', 'background', 'wind', 'combined']):

        if var in ['retrieval', 'background', 'wind']:
            d = (ts[var] - ts["method"]).est_emissions
        else:
            d = (ts[var].est_emissions -
                 ts["method"].true_emissions.loc[ts["method"].overpass])

        mb = np.nanmean(d.values)
        sd = np.nanstd(d.values)
        n = np.sum(np.isfinite(d.values))

        mean_bias[var] = mb
        scatter[var] = sd
        numbers[var] = n

        if show:
            text = 'MB: %.1f Mt yr$^{-1}$ (%d %%)' % (mb, 100 * mb / 20.0)
            axes[i].text(-29.2, 14.4, text, va='top', fontsize='small')

            text = 'SD: %.1f Mt yr$^{-1}$ (%d %%)' % (sd, 100 * sd / 20.0)
            axes[i].text(-29.2, 13.2, text, va='top', fontsize='small')

            text = 'N: %d' % n
            axes[i].text(-29.2, 11.8, text, va='top', fontsize='small')

            axes[i].hist(d, bins=np.linspace(-30,30,31))

            if i == 2:
                axes[i].set_xlabel('difference (Mt yr$^{-1}$)')

            axes[i].set_xlim(-30,30)
            axes[i].set_xticks(np.arange(-30,31,10))
            axes[i].set_ylim(0, 15)

            if var == 'combined':
                title = 'Total'
            else:
                title = var.capitalize()

            axes[i].set_title('(%s) %s error' % ('abcde'[i], title))

            if i == 0:
                axes[i].set_ylabel('number of estimates')


    if show:
        plt.tight_layout()

    return mean_bias, scatter, numbers


def plot_mb_bars():

    mbs = defaultdict(list)
    sds = defaultdict(list)

    for no2_ssp in ['none', 'high']:
        for co2_ssp in ['low', 'medium', 'high']:
            for method in ['sub-areas', 'gauss']:
                mb, sd, n = plot_mb_histograms(method=method, co2_ssp=co2_ssp,
                                               no2_ssp=no2_ssp, show=False)

                for var in ['method', 'retrieval', 'background', 'wind', 'combined']:

                    if no2_ssp == 'none':
                        name = 'co2-%s' % var
                    else:
                        name = 'no2-%s' % var

                    mbs[name].append( mb[var] )
                    sds[name].append( sd[var] )

    return mbs, sds



def plot_mb_bars2(mean_bias, scatter, no2_ssp='none'):

    x = np.arange(3.0)

    fig, axes = plt.subplots(2, 5, figsize=(9, 4.5))

    i = 0
    for var in ['method', 'retrieval', 'background', 'wind', 'combined']:

            if no2_ssp == 'none':
                name = 'co2-%s' % var
            else:
                name = 'no2-%s' % var

            mb = 100.0 * np.array(mean_bias[name]) / 20.0
            sd = 100.0 * np.array(scatter[name]) / 20.0

            axes[0,i].bar(x, mb[0::2], align='edge', width=-0.4,
                          label='Sub-polygons')
            axes[0,i].bar(x, mb[1::2], align='edge', width=+0.4,
                          label='Gaussian')

            axes[1,i].bar(x, sd[0::2], align='edge', width=-0.4,
                          label='Sub-polygons')
            axes[1,i].bar(x, sd[1::2], align='edge', width=+0.4,
                          label='Gaussian')

            axes[0,i].set_xticks(x)
            axes[1,i].set_xticks(x)
            axes[0,i].set_xticklabels(['0.5', '0.7', '1.0'])
            axes[1,i].set_xticklabels(['0.5', '0.7', '1.0'])

            axes[0,i].set_yticks(np.arange(-50,51,5), minor=True)
            axes[1,i].set_yticks(np.arange(0,101,5), minor=True)

            axes[0,i].set_ylim(-50,50)
            axes[1,i].set_ylim(0, 100)


            axes[0,i].grid(True, axis='y')
            axes[1,i].grid(True, axis='y')

            if var == 'combined':
                title = 'Total'
            else:
                title = var.capitalize()

            axes[0,i].set_title('(%s) %s' % ('abcdefghijkl'[i],
                                             title))

            if i > 0:
                axes[0,i].yaxis.set_ticklabels([])
                axes[1,i].yaxis.set_ticklabels([])

            i += 1


    axes[0,0].set_ylabel('Mean bias (%)')
    axes[1,0].set_ylabel('Standard deviation (%)')
    axes[1,2].set_xlabel('$\sigma_\mathrm{VEG50}$ (ppm)')

    fig.tight_layout()
    fig.align_labels()

    axes[0,0].legend(loc='upper left', fontsize='small')










def compute_true_emissions(emissions, overpass, hours=None):

    if hours is None:
        hours = [0]

    true_emissions = np.mean([
            emissions.loc[overpass - pandas.Timedelta(h, unit='hour')].values
            for h in hours
    ], axis=0)
    return true_emissions



def boxplot(datasets, nsat='6', nmax=None, sample='same', indx=None,
            origin='Berlin', use_constant=False, vmin=-20, vmax=20,
            is_mce=False, hours=None, show_header=True):
    """
    """

    # filter fo nsat
    datasets = [filter_nsat(d, nsat) for d in datasets]

    data = []
    unc = []
    overpasses = []

    for d in datasets:
        true_emissions = compute_true_emissions(d['true_emissions'],
                                                d.overpass, hours=hours)

        y = d['est_emissions'] - true_emissions
        u = d['est_emissions_std']

        valids = np.isfinite(y) #& np.isfinite(u)

        overpasses.append(d.overpass[valids].values)

        data.append(y[valids])
        unc.append(u[valids])


    if sample == 'same':
        ind = sorted(set(overpasses[0]) & set(overpasses[1]) &
                     set(overpasses[2]))

        nn = [len(v) for v in data]
        data = [d.loc[ind] for d in data]
    else:
        nn = None

    if indx is not None:
        data = [d.loc[indx] for d in data]

    sep = ' & '
    line = sep.join(['noise', '#plumes', 'MB', 'MB', 'SD', 'SD'])
    if show_header:
        print(line)


    for i, (noise, d,u) in enumerate(zip([0.5, 0.7, 1.0], data, unc)):

        # annual mean emissions at overpass used for computing relative errors
        if origin == 'Berlin':
            if use_constant:
                t = 100.0 / 16.8
            else:
                t = 100 / 20.0 # em_mean.mean()
        elif origin == 'Janschwalde':
            t = 100 / 40.0 # Janschwalde (at overpass)
        else:
            raise ValueError

        n = len(d)
        mb  = np.mean(d)
        std = np.std(d)

        print("", end=sep)
        print('%.1f' % noise, end=sep)
        print('%s / %s' % (n, nn[i]), end=sep)
        print('%5.1f%s%5.1f' % (mb, sep, mb*t), end=sep)
        print('%5.1f%s%5.1f \\\\' % (std, sep, std*t))

    #print('Mean emissions: %.2f Mt/yr' % 20.0) #em_mean.mean())
    print(r"\middlehline")

    fig = plt.figure(figsize=(3.3,3.5))
    ax = fig.add_subplot(111)

    for i,n in enumerate(data):
        if nn is None:
            ax.text(i+1,vmax+0.1, str(len(n)), ha='center', va='bottom')
        else:
            text = '%s/%s' % (len(n), nn[i])
            ax.text(i+1,vmax+0.1, text, ha='center', va='bottom')


    ax.boxplot(data, showfliers=True, whis=[5,95], bootstrap=5000, widths=0.4,
               labels=['0.5', '0.7', '1.0'], meanline=True, showmeans=True)

    ax.set_xlabel('$\sigma_{VEG50}$ (ppm)')
    ax.set_ylabel('Deviation (Mt yr$^{-1}$)')

    ax.set_ylim(vmin,vmax)
    ax.grid(True, ls='-', color='gray')
    plt.tight_layout(pad=0.25)
    fig.subplots_adjust(top=0.95)

    return data




def plot_wind_error():
    """
    Plot two plots showing wind speed and wind error (difference between
    500 m and eff wind speed)
    """
    fig, axes = plt.subplots(nrows=2, figsize=(9,4))

    ts_method = io.read_mass_balance('CO2', which='method',
                                     method='sub-areas')
    ts_wind = io.read_mass_balance('CO2', which='wind',
                                   method='sub-areas')

    axes[0].plot(ts_wind.overpass, ts_wind.wind_speed_eff, 'o', mfc='none',
                 label='effective wind speed')

    axes[0].plot(ts_wind.overpass, ts_wind.wind_speed_500m, 's', mfc='none',
                 label='1-km wind speed')


    axes[0].set_title('(a)')
    axes[0].set_ylim(0,15)
    axes[0].set_ylabel('wind speed (m s$^{-1}$)')
    axes[0].set_xlim('2015-01-01', '2016-01-01')
    axes[0].grid(True)
    axes[0].legend(ncol=2)



    d_wind = ts_wind.wind_speed_500m - ts_method.wind_speed_eff
    d_emis = ts_wind.est_emissions - ts_method.est_emissions

    valids = np.isfinite(d_wind) & np.isfinite(d_emis)
    d_wind = d_wind[valids]
    d_emis = d_emis[valids]


    lines = axes[1].plot(d_wind.overpass, d_wind, 'o',
                         label='wind', mfc='none')

    axes[1].set_title('(b)')
    axes[1].set_ylim(-8, 8)
    axes[1].set_ylabel('wind error (m s$^{-1}$)')
    axes[1].set_xlim('2015-01-01', '2016-01-01')
    axes[1].grid(True)

    ax3 = axes[1].twinx()
    lines += ax3.plot(d_emis.overpass, d_emis, 'rD',
                      label='emissions', mfc='none')

    ax3.set_ylim(-32, 32)
    ax3.set_yticks([-20, 0, 20])
    ax3.set_ylabel('wind error (Mt yr$^{-1}$)')

    ax3.legend(lines, [l.get_label() for l in lines], ncol=2)

    plt.tight_layout()
    fig.align_ylabels()

    mb = np.nanmean(d_wind)
    sd = np.nanstd(d_wind)

    print(mb, sd)

    re = scipy.stats.linregress(d_wind, d_emis)

    print(re)





def model_uncertainty(filename, nsat='6', sigma=1.0):
    """
    Model uncertainty depending on number of pixels and wind speed. The model
    does not include signal strength.
    """
    ts = io.read_results(filename, nsat=nsat)

    n_pixels = ts['number of pixels'].values
    unc = ts['CO2 uncertainty (Mt/yr)'].values

    wind_speed = np.array([emiss.gauss.get_wind_speed(t)['mean wind (m/s)']
                           for t in ts.index])


    valids = np.isfinite(unc) & (unc < 50.0)

    wind_speed = wind_speed[valids]
    n_pixels = n_pixels[valids]
    unc = unc[valids]

    inv_n = 1.0 / np.sqrt(n_pixels)

    # model
    X = np.transpose([sigma * inv_n, wind_speed])
    #X = sm.add_constant(X)
    res = sm.RLM(unc, X).fit()

    print(res.summary())

    slope1, slope2 = res.params


    fig, axes = plt.subplots(2,1, figsize=(6,6))

    # number of pixels
    axes[0].plot(inv_n, unc, 'o',
                 label='Uncertainty if less than 50 Mt yr$^{-1}$')
    axes[0].plot(inv_n, slope1 * sigma * inv_n,
                 label='Robust linear model (m = %2.1f)' % slope1)

    axes[0].set_xlabel('Number of pixels', labelpad=-15)

    n = np.array([10000, 1000, 100, 10])
    axes[0].set_xticks(1.0/np.sqrt(n))
    axes[0].set_xticklabels([str(i) for i in n], rotation=45)
    axes[0].set_xlim(0, 1.0/np.sqrt(10))

    # wind speed
    axes[1].plot(wind_speed, unc, 'o',
                 label='Uncertainty if less than 50 Mt yr$^{-1}$')
    axes[1].plot(wind_speed, slope2 * wind_speed,
                 label='Robust linear model (m = %.2g)' % slope2)

    axes[1].set_xlabel('Wind speed (m s$^{-1}$)')
    axes[1].set_xlim(0,14)

    for i, ax in enumerate(axes):
        ax.set_title('(%s)' % 'abc'[i])
        ax.set_ylim(-10,50)
        ax.set_ylabel('CO$_2$ uncertainty (Mt yr$^{-1}$)')
        ax.grid(True)
        ax.legend(loc=2)

    plt.tight_layout(pad=0.2, h_pad=0.4)

    return res




def show_overpass_orbits():

    import smartcarb.config
    import smartcarb.misc
    import smartcarb.io
    import amrs.vis

    from string import ascii_lowercase

    filenames = [
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015010111_o0000_l0000.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015010211_o0015_l0000.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015010311_o0030_l0401.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015010411_o0045_l0805.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015010511_o0060_l0805.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015010611_o0075_l1207.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015010711_o0090_l1207.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015010811_o0105_l1610.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015010911_o0120_l1610.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015011011_o0135_l2012.nc',
         '/output/SMARTCARB/output/level2/Sentinel_7_CO2/Sentinel_7_CO2_2015011111_o0150_l2012.nc',
    ]
    for i, fn in enumerate(filenames):
        d = smartcarb.io.read_level2(fn)
        amrs.vis.make_level2_map(d.lonc, d.latc, np.ones_like(d.clouds),
                                 domain=smartcarb.config.DOMAINS['smartcarb'],
                                 fig_size=2, add_colorbar=False)

        c = 'abcdef'[ list(smartcarb.misc.lon_eq_by_nsat('6')).index(d.lon_eq) ]
        plt.gca().set_title('(%s) %s' % (ascii_lowercase[i], c))





def plot_cloud_fraction_spectrum():
    """
    Plot a spectrum.
    """
    filename = "/output/SMARTCARB/output/timeseries/ts_cosmo_modis_cloud_mask.dat"
    cc = pandas.read_csv(filename, index_col=0, parse_dates=True)

    plt.figure()

    for name in ['COSMO', 'MODIS']:
        freqs, psd = scipy.signal.welch(cc['%s clear fraction' % name])
        plt.plot(1/freqs, psd, '-', label=name)

    plt.xscale('log', basex=2)
    plt.xticks([2,4,8,16,32,64,128,256], ['2','4','8','16','32','64','128','256'] )
    plt.xlabel('days')
    plt.ylabel('power')
    plt.legend()
    plt.ylim(0,1.2)
    plt.grid(True, which='both')



def plot_across_values(data, tracer, xa=10e3, xb=20e3, ya=-30e3, yb=30e3):
    _, x, y, xc, yc, c, c_std, detection, wind = emiss.mass.extract_pixels(
        data, tracer, xa, xb, ya, yb
    )

    figsize = np.array(plt.rcParams['figure.figsize']) * 2 / 3
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes([0.15,0.17,0.82,0.8])

    ax.plot(y[detection]/1e3, c[detection], 'o', color='tab:red',
            label='Inside detected plume')
    ax.plot(y[~detection]/1e3, c[~detection], 'o', mfc='none', color='tab:red',
            label='Outside detected plume')


    # background
    _, x, y, xc, yc, c, c_std, detection, wind = emiss.mass.extract_pixels(
        data, 'CO2_estimated_background', xa, xb, ya, yb
    )
    sort = y.argsort()
    y = y[sort]
    c = c[sort]
    mask = np.isfinite(c)

    ax.plot(y[mask]/1e3, c[mask], '--', color='tab:blue',
            label='Estimated background')

    _, x, y, xc, yc, c, c_std, detection, wind = emiss.mass.extract_pixels(
        data, 'CO2_true_background', xa, xb, ya, yb
    )
    sort = y.argsort()
    y = y[sort]
    c = c[sort]
    mask = np.isfinite(c)
    ax.plot(y[mask]/1e3, c[mask], '-', color='tab:blue',
            label='True background')


    ax.set_xlabel('Across-plume direction (km)')
    ax.set_ylabel('XCO$_2$ (ppm)')

    ax.set_ylim(404, 408)
    ax.set_xlim(ya/1e3, yb/1e3)
    ax.grid(True)

    plt.legend(loc=2, fontsize='small')







