#!/usr/bin/env python
# coding: utf-8

"""
Extract true emissions for Berlin and power plants.
"""

import os

from datetime import datetime, timedelta
from netCDF4 import Dataset

import numpy as np

import smartcarb as sc
import amrs.misc.transform as transform



# area of COSMO grid cell
AREA = 1235941.1 # in m2



def main(start, end, out_filename, emiss_path, tracer='CO2'):
    """
    Extract emission of Berlin and point souces. For January and July,
    correction factors are applied to correct industry emissions.
    """

    file_exists = os.path.exists(out_filename)
    mode = 'a' if file_exists else 'w'

    facilities = sorted(sc.config.POINT_SOURCES.keys())


    with open(out_filename, mode) as outfile:

        if not file_exists:
            outfile.write('\t'.join(
                ['UTC', 'Berlin (Mt/yr)'] + ['%s (Mt/y)' % f for f in facilities]
            ))
            outfile.write('\n')

        for time in sc.misc.iter_times(start, end, timedelta(hours=1)):
            print(time)

            pattern = 'CO2_CO_NOX_Berlin-big_%Y%m%d%H.nc'
            filename = os.path.join(emiss_path, time.strftime(pattern))

            with Dataset(filename) as nc:

                if tracer == 'CO2':
                    b0 = np.sum(nc.variables['CO2_B0_E'][:]) * AREA # kg s-1
                    b1 = np.sum(nc.variables['CO2_B1_E'][:]) * AREA # kg s-1
                    b2 = np.sum(nc.variables['CO2_B2_E'][:]) * AREA # kg s-1
                else:
                    b0 = np.sum(nc.variables['NOX_B0_E'][:]) * AREA # kg s-1
                    b1 = np.sum(nc.variables['NOX_B1_E'][:]) * AREA # kg s-1
                    b2 = np.sum(nc.variables['NOX_B2_E'][:]) * AREA # kg s-1

                # fix emissions for january and july
                if time.month in [1,7]:
                    berlin = 0.55 * b0 + b1 + b2
                else:
                    berlin = b0 + b1 + b2

                berlin *= 0.031536 # kg/s -> Mt/y

                # write time and berlin emissions
                outfile.write(str(time))
                outfile.write('\t')
                outfile.write('%e' % berlin)


                # power stations
                if time == start:
                    rlon = nc.variables['rlon'][:]
                    rlat = nc.variables['rlat'][:]

                    rlon, rlat = np.meshgrid(rlon, rlat)
                    lon, lat = transform.rotpole2wgs(rlon, rlat, sc.config.pollon,
                                                     sc.config.pollat)

                if tracer == 'CO2':
                    var = nc.variables['CO2_JV_E']
                else:
                    var = nc.variables['NOX_JV_E']

                for name in facilities:

                    lon_o, lat_o = sc.config.POINT_SOURCES[name]
                    i,j = sc.misc.nearest_index(lon_o, lat_o, lon, lat)

                    v = np.sum(var[:,:,i-2:i+3,j-2:j+3]) * AREA # kg s-1
                    v *= 0.031536 # kg/s -> Mt/y

                    outfile.write('\t')
                    outfile.write('%e' % v)


            outfile.write('\n')



if __name__ == '__main__':
    tracer = 'NO2'

    out_filename = 'ts_%s_emissions.dat' % tracer

    emiss_path = '/project/smartcarb/emissions/2015/'
    emiss_path = '/project/brd134/smartcarb/emissions/2015'
    start = datetime(2015,1,1)
    end = datetime(2016,1,1)

    main(start, end, out_filename, tracer=tracer, emiss_path=emiss_path)





